import re
from shutil import move, copy
from pathlib import Path
from os.path import relpath
from uuid import uuid1

from foliant.preprocessors.base import BasePreprocessor


class Preprocessor(BasePreprocessor):
    defaults = {'mkdocs_project_dir_name': 'mkdocs'}

    _image_pattern = re.compile(r'\!\[(?P<caption>.*)\]\((?P<path>((?!:\/\/).)+)\)')

    def _collect_images(self, content: str, md_file_path: Path) -> str:
        '''Find images outside the source directory, copy them into the source directory,
        and replace the paths in the source.

        This is necessary because MkDocs can't deal with images outside the project's doc_dir.

        :param content: Markdown content
        :param md_file_path: Path to the Markdown file with content ``content``

        :returns: Markdown content with image paths pointing within the source directory
        '''

        self.logger.debug(f'Looking for images in {md_file_path}.')

        def _sub(image):
            image_caption = image.group('caption')
            image_path = (md_file_path.parent / Path(image.group('path'))).resolve()

            self.logger.debug(f'Detected image: caption="{image_caption}", path={image_path}')

            if self.working_dir.resolve() not in image_path.parents:
                self.logger.debug('Image outside source directory.')

                self._collected_imgs_path.mkdir(exist_ok=True)

                collected_img_path = (
                    self._collected_imgs_path/f'{image_path.stem}_{str(uuid1())}'
                ).with_suffix(image_path.suffix)

                copy(image_path, collected_img_path)

                self.logger.debug(f'Image copied to {collected_img_path}')

                rel_img_path = Path(relpath(collected_img_path, md_file_path.parent)).as_posix()

            else:
                self.logger.debug('Image inside source directory.')
                rel_img_path = Path(relpath(image_path, md_file_path.parent)).as_posix()

            img_ref = f'![{image_caption}]({rel_img_path})'

            self.logger.debug(f'Replacing with: {img_ref}')

            return img_ref

        return self._image_pattern.sub(_sub, content)


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._collected_imgs_path = self.working_dir / f'_img_{str(uuid1())}'

        self.logger = self.logger.getChild('mkdocs')

        self.logger.debug(f'Preprocessor inited: {self.__dict__}')

    def apply(self):
        for markdown_file_path in self.working_dir.rglob('*.md'):
            with open(markdown_file_path, encoding='utf8') as markdown_file:
                content = markdown_file.read()

            processed_content = self._collect_images(content, markdown_file_path)

            if processed_content:
                with open(markdown_file_path, 'w', encoding='utf8') as markdown_file:
                    markdown_file.write(processed_content)

        mkdocs_project_path = self.working_dir / self.options['mkdocs_project_dir_name']
        mkdocs_tmp_project_path = self.working_dir / str(uuid1())
        mkdocs_src_path = mkdocs_tmp_project_path / 'docs'
        mkdocs_src_path.mkdir(parents=True)

        for item in self.working_dir.glob('*'):
            if item.name == mkdocs_tmp_project_path.name:
                continue
            else:
                move(str(item), str(mkdocs_src_path))

        move(str(mkdocs_tmp_project_path), str(mkdocs_project_path))

        self.logger.debug(f'Moved {mkdocs_tmp_project_path} to {mkdocs_project_path}')



import re
from shutil import rmtree, copytree
from pathlib import Path
from typing import Dict
from subprocess import run, PIPE, STDOUT, CalledProcessError

from yaml import dump

from foliant.utils import spinner
from foliant.backends.base import BaseBackend


class Backend(BaseBackend):
    targets = ('site', 'mkdocs', 'ghp')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._mkdocs_config = self.config.get('backend_config', {}).get('mkdocs', {})

        self._mkdocs_site_dir_name = f'{self._mkdocs_config.get("slug", self.get_slug())}.mkdocs'

        self._mkdocs_project_dir_name = f'{self._mkdocs_site_dir_name}.src'

        self.required_preprocessors_after = {
            'mkdocs': {
                'mkdocs_project_dir_name': self._mkdocs_project_dir_name
            }
        },

        self.logger = self.logger.getChild('mkdocs')

        self.logger.debug(f'Backend inited: {self.__dict__}')

    def _escape_control_characters(self, source_string: str) -> str:
        '''Escape control characters such as double quotation mark, dollar sign, backtick
        to use them in system shell commands.

        :param source_string: String that contains control characters

        :returns: String with control characters escaped by backslash
        '''

        escaped_string = source_string.replace('"', "\\\"").replace('$', "\\$").replace('`', "\\`")

        return escaped_string

    def _get_build_command(self, mkdocs_site_path: Path) -> str:
        '''Generate ``mkdocs build`` command to build the site.

        :param mkdocs_site_path: Path to the output directory for the site
        '''

        components = [self._mkdocs_config.get('mkdocs_path', 'mkdocs')]
        components.append('build')
        components.append(f'-d "{self._escape_control_characters(str(mkdocs_site_path))}"')

        command = ' '.join(components)

        self.logger.debug(f'Build command: {command}')

        return command

    def _get_ghp_command(self) -> str:
        '''Generate ``mkdocs gh-deploy`` command to deploy the site to GitHub Pages.'''

        components = [self._mkdocs_config.get('mkdocs_path', 'mkdocs')]
        components.append('gh-deploy')

        command = ' '.join(components)

        self.logger.debug(f'GHP upload command: {command}')

        return command

    def _get_page_with_optional_heading(self, page_file_path: str) -> str or Dict:
        '''Get the content of first heading of source Markdown file, if the file
        contains any headings. Return a data element of ``pages`` section
        of ``mkdocs.yml`` file.

        :param page_file_path: path to source Markdown file

        :returns: Unchanged file path or a dictionary: content of first heading, file path
        '''

        self.logger.debug(f'Looking for the first heading in {page_file_path}')

        if page_file_path.endswith('.md'):
            page_file_full_path = self.working_dir /\
                self._mkdocs_project_dir_name / 'docs' / page_file_path

            with open(page_file_full_path, encoding='utf8') as page_file:
                content = page_file.read()

                headings_found = re.search(
                    r'^\#{1,6}\s+(.+?)(?:\s+\{\#\S+\})?\s*$',
                    content,
                    flags=re.MULTILINE
                )

                if headings_found:
                    first_heading = headings_found.group(1)
                    self.logger.debug(f'Heading found: {first_heading}')
                    return {first_heading: page_file_path}

        self.logger.debug(f'No heading found, returning original file path.')

        return page_file_path

    def _get_pages_with_headings(self, pages: Dict) -> Dict:
        '''Update ``pages`` section of ``mkdocs.yml`` file with the content
        of top-level headings of source Markdown files.

        param pages: Dictionary with the data of ``pages`` section

        returns: Updated dictionary
        '''

        def _recursive_process_pages(pages_subset, parent_is_dict):
            if isinstance(pages_subset, dict):
                new_pages_subset = {}
                for key, value in pages_subset.items():
                    if not key:
                        key = self._mkdocs_config.get('default_subsection_title', '…')

                    new_pages_subset[key] = _recursive_process_pages(value, True)

            elif isinstance(pages_subset, list):
                new_pages_subset = []
                for item in pages_subset:
                    new_pages_subset.append(_recursive_process_pages(item, False))

            elif isinstance(pages_subset, str):
                if not parent_is_dict:
                    new_pages_subset = self._get_page_with_optional_heading(pages_subset)

                else:
                    new_pages_subset = pages_subset

            else:
                new_pages_subset = pages_subset

            return new_pages_subset

        new_pages = _recursive_process_pages(pages, False)

        self.logger.debug(f'All pages with their headings: {new_pages}')

        return new_pages

    def make(self, target: str) -> str:
        with spinner(f'Making {target} with MkDocs', self.logger, self.quiet, self.debug):
            try:
                mkdocs_project_path = self.working_dir / self._mkdocs_project_dir_name

                config = self._mkdocs_config.get('mkdocs.yml', {})

                self.logger.debug(f'Backend config: {config}')

                if 'site_name' not in config and self._mkdocs_config.get('use_title', True):
                    config['site_name'] = self.config['title']

                if 'pages' not in config and self._mkdocs_config.get('use_chapters', True):
                    config['pages'] = self.config['chapters']

                if self._mkdocs_config.get('use_headings', True):
                    config['pages'] = self._get_pages_with_headings(config['pages'])

                self.logger.debug(f'mkdocs.yml: {config}')

                with open(mkdocs_project_path/'mkdocs.yml', 'w', encoding='utf8') as mkdocs_config:
                    self.logger.debug(f'Saving mkdocs.yml into {mkdocs_project_path}')
                    dump(
                        config,
                        mkdocs_config,
                        default_flow_style=False
                    )

                if target == 'mkdocs':
                    rmtree(self._mkdocs_project_dir_name, ignore_errors=True)
                    copytree(mkdocs_project_path, self._mkdocs_project_dir_name)

                    return self._mkdocs_project_dir_name

                elif target == 'site':
                    try:
                        mkdocs_site_path = Path(self._mkdocs_site_dir_name).absolute()
                        run(
                            self._get_build_command(mkdocs_site_path),
                            shell=True,
                            check=True,
                            stdout=PIPE,
                            stderr=STDOUT,
                            cwd=mkdocs_project_path
                        )

                        return self._mkdocs_site_dir_name

                    except CalledProcessError as exception:
                        raise RuntimeError(f'Build failed: {exception.output.decode()}')

                elif target == 'ghp':
                    try:
                        mkdocs_site_path = Path(self._mkdocs_site_dir_name).absolute()
                        process = run(
                            self._get_ghp_command(),
                            shell=True,
                            check=True,
                            stdout=PIPE,
                            stderr=STDOUT,
                            cwd=mkdocs_project_path
                        )
                        ghp_url = process.stdout.decode().splitlines()[-1].split(': ')[-1]

                        return ghp_url

                    except CalledProcessError as exception:
                        raise RuntimeError(
                            f'GitHub Pages deploy failed: {exception.output.decode()}'
                        )

                else:
                    raise ValueError(f'MkDocs cannot make {target}')

            except Exception as exception:
                raise type(exception)(f'Build failed: {exception}')



from setuptools import setup


SHORT_DESCRIPTION = 'MkDocs backend for Foliant documentation generator.'

try:
    with open('README.md', encoding='utf8') as readme:
        LONG_DESCRIPTION = readme.read()

except FileNotFoundError:
    LONG_DESCRIPTION = SHORT_DESCRIPTION


setup(
    name='foliantcontrib.mkdocs',
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    version='1.0.11',
    author='Konstantin Molchanov',
    author_email='moigagoo@live.com',
    url='https://github.com/foliant-docs/foliantcontrib.mkdocs',
    packages=['foliant.backends', 'foliant.preprocessors'],
    license='MIT',
    platforms='any',
    install_requires=[
        'foliant>=1.0.8',
        'mkdocs>=1.0.4'
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Documentation",
        "Topic :: Utilities",
    ]
)
